'------------------------------------------------------------------------------
' 
' 	BSD License
'   -----------
' 
'   Copyright (c) 2013 Elba Corp.  All rights reserved.
' 
'      Elba Corp.
'      10915 SW Avocet Ct.
'      Beaverton, OR  97007
' 
'   Redistribution and use in source and binary forms, with or without
'   modification, are permitted only for use by licensees of the ZBasic
'   Compiler for generic AVR devices and only then if all the following
'   conditions are met:
' 
'   - Redistributions of source code must retain the above copyright notice,
'     this list of conditions and the following disclaimer.
' 
'   - Redistributions in binary form must reproduce the above copyright notice,
'     this list of conditions and the following disclaimer in the documentation
'     and/or other materials provided with the distribution.
' 
'   - Neither the name of the copyright owner nor the names of its contributors
'     may be used to endorse or promote products derived from this software
'     without specific prior written permission of the copyright owner.
' 
'   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
'   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
'   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
'   ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
'   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
'   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
'   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
'   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
'   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
'   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
'   POSSIBILITY OF SUCH DAMAGE.
' 
'------------------------------------------------------------------------------

'------------------------------------------------------------------------------
'
' This module demonstrates how to use the Arduino Ethernet classes to request
' the current time from an NTP server.  If enabled, the test driver sets the
' ZBasic RTC using the received data, adjusted for local time.
'
' There are several configurable elements (unless otherwise indicated, found
' in this file):
'
'    - the ZBasic SPI channel to use (specified in in Arduino/SPI/SPI.h)
'    - the SPI chip select pin of the W5100 chip
'    - the MAC address for the W5100 chip
'    - the name or IP address of the NTP server to use
'    - the offset from GMT to the local time zone
'    - if daylight savings time should be observed
'
' Although it shouldn't need to be changed, the client port address to use
' can also be modified.
' 
'------------------------------------------------------------------------------

' uncomment the following line to include a Main() test driver
#define INCLUDE_TEST_DRIVER

' uncomment the line below to use a local NTP time server (specify IP address below)
'#define USE_LOCAL_NTP_SERVER

' allow Functions to be invoked as if they are Subroutines
Option CallFunctions

#if defined(INCLUDE_TEST_DRIVER)
' specify Arduino compatibility
Option Arduino "C:/Projects/Arduino"
#endif

' import the necessary Arduino Ethernet declarations
#import Public "NetTimeImport.h" "#NET_TIME_DHCP", "#NET_TIME_DNS"

' specify the CS pin for the W5100 chip
Private Const csPin as Byte = 14

' specify the MAC address of the W5100 chip (must be unique on a local network)
Private myMacAddr as ByteVectorData ({ &H00, &H10, &H5a, &H63, &H2e, &H85 })

#if Not defined(NET_TIME_DHCP)
	' specify the host IP address and gateway IP address
	Dim ipAddr as IPAddress = _Create(192, 168, 0, 31)
	Dim gwAddr as IPAddress = _Create(192, 168, 0, 2)
  #if Defined(NET_TIME_DNS)
	' specify the DNS server IP address
	Dim dnsAddr as IPAddress =  _Create(192, 168, 0, 2)
  #endif
#endif

' specify the IP address of the NTP time server
#if defined(NET_TIME_DNS)
	Private Const timeServer as String = "us.pool.ntp.org"
'	Private Const timeServer as String = "clepsydra.dec.com"
#elseif defined(USE_LOCAL_NTP_SERVER)
	Private timeServer as IPAddress = _Create(192, 168, 0, 1)
#else
	Private timeServer as IPAddress = _Create(69, 167, 160, 102)	' us.pool.ntp.org
'	Private timeServer as IPAddress = _Create(204, 123, 2, 5)		' clepsydra.dec.com
#endif

' specify the local time zone relative to GMT
Private Const tzAdjMinutes as Integer = CInt(-8 * MinutesPerHour)

' indicate if daylight savings time algorithm should be used - see dstInEffect()
Private Const useDST as Boolean = True

' specify the local port to be used and the standard NTP port
Private Const localPort as UnsignedInteger = 8888	' local port to listen for UDP packets
Private Const sntpPort as UnsignedInteger = 123		' standard SNTP port

'-------------------------------------------------------------------------------
' conversion constants
Private Const tzAdjustSeconds as UnsignedLong = CLng(tzAdjMinutes * CInt(SecondsPerMinute))
Private Const SecondsPerMinute as UnsignedLong = 60
Private Const MinutesPerHour as UnsignedLong = 60
Private Const HoursPerDay as UnsignedLong = 24
Private Const SecondsPerHour as UnsignedLong = MinutesPerHour * SecondsPerMinute
Private Const SecondsPerDay as UnsignedLong = HoursPerDay * SecondsPerHour
Private Const DeltaZBasicNTP as UnsignedLong = 3124137600	' seconds btw 1900/1/1 and 1999/1/1

'-------------------------------------------------------------------------------
' These structure represent the basic NTP information (See RFC 5905 section 6).
Private Structure TimePacketNTP
	Dim seconds as UnsignedLong		' seconds since 00:00, 01 Jan 1900
	Dim fraction as UnsignedLong	' fraction of a second
End Structure

' This structure corresponds to the basic Network Time Protocol data.
' See RFC 5905 section 7.3 for more information.
Private Structure BasicInfoNTP
	Dim LiVnMode as Byte
	Dim stratum as Byte
	Dim poll as Byte
	Dim precision as Byte
	Dim rootDelay as Long
	Dim rootDispersion as Long
	Dim referenceID(1 to 4) as Byte
	Dim referenceTime as TimePacketNTP
	Dim originateTime as TimePacketNTP
	Dim receiveTime as TimePacketNTP
	Dim transmitTime as TimePacketNTP	' NTP server time returned here
End Structure

' This Program Memory array gives the three-letter abbreviation
' for months numbered 1 to 12.
Private monthName as StringVectorData ({
	"Jan", "Feb", "Mar", "Apr", "May", "Jun"
	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
})

'-------------------------------------------------------------------------------
'
'' GetTimeNTP
'
' Get the current time, adjusted for timezone and DST.  The return value represents
' seconds since 1900/01/01.  If the time cannot be obtained from the NTP server
' the return value is zero.
'
' N.B.: This function uses the RTC stopwatch to detect a timeout condition.
'
Public Function GetTimeNTP(ByRef eth as EthernetBasicClass) as UnsignedLong
	getTimeNTP = 0
	Dim udpObj as EthernetBasicUDP

	Call udpObj.begin(localPort)

	' begin sending the NTP packet
#if defined(NET_TIME_DNS)
	Dim serverIP as IPAddress
	If (Not getHostAddress(eth, timeServer, serverIP)) Then
		Debug.Print "failed to resolve host name """; timeServer; """";
		Exit Function
	End If
  #if 0
	Debug.Print "resolved """; timeServer; """ to ";
	Call dspIpAddr(serverIP, True)
  #endif
	Call udpObj.beginPacket(serverIP, sntpPort)
#else
	Call udpObj.beginPacket(timeServer, sntpPort)
#endif

	' send the rest of the request
	Dim ntp as BasicInfoNTP
	Call MemSet(@ntp, SizeOf(ntp), 0)
	ntp.LiVnMode = &H1b	' ver 3, mode 3 (client)
	Dim buf() as Byte Based @ntp
	Call udpObj.write(buf, SizeOf(ntp))
	Call udpObj.endPacket()

	' wait for a reply from the server observing a timeout
	Dim packetSize as Integer
	Register.RTCStopWatch = 0
	Do
		packetSize = udpObj.parsePacket()
		If (packetSize > 0) Then
			Exit Do
		End If
		Call Sleep(0.5)
	Loop While (Register.RTCStopWatch < CUInt(5.0 * CSng(Register.RTCTickFrequency)))

	' retrieve the reply packet and process it
	Const MinMsgSize as Integer = Sizeof(ntp) + 8	' packet size plus UDP overhead
	If (packetSize > 0) Then
		' get the NTP payload
		Call udpObj.read(buf, SizeOf(ntp))

		' convert to little endian form and adjust to ZBasic epoch and local time zone
		Dim seconds as UnsignedLong
		seconds = reverseBytes(ntp.transmitTime.seconds) - DeltaZBasicNTP + tzAdjustSeconds

		' determine the day number (relative to 1/1/1999)
		Dim day as UnsignedInteger
		day = CUInt(seconds \ SecondsPerDay)
		seconds = seconds Mod SecondsPerDay
		
		' adjust the time for DST, if applicable
		If (useDST And dstInEffect(day, seconds)) Then
			seconds = seconds + SecondsPerHour
			If (seconds >= SecondsPerDay) Then
				seconds = seconds - SecondsPerDay
				day = day + 1
			End If
		End If

		' convert back to seconds since 1900/01/01
		GetTimeNTP = seconds + (CULng(day) * SecondsPerDay) + DeltaZBasicNTP
	End If

	Call udpObj.stop()
End Function

'-------------------------------------------------------------------------------
'
'' reverseBytes
'
' Reverse the order of bytes in a 32-bit value.  This is used to convert
' from host byte order (little endian) to network byte order (big endian).
'
#pragma warning(push; 2:off)
Private Function reverseBytes(ByRef ulval as UnsignedLong) as UnsignedLong
	Dim src() as Byte Based @ulval
	Dim dst() as Byte Based reverseBytes.DataAddress

	dst(1) = src(4)
	dst(2) = src(3)
	dst(3) = src(2)
	dst(4) = src(1)
End Function
#pragma warning(pop)

'-------------------------------------------------------------------------------
'
'' dstInEffect
'
' Determine if Daylight Saving Time is in effect as of the ZBasic day number and
' second count specified.  The start/stop dates are the second Sunday in March 
' and the first Sunday of November (i.e. the 2007 U.S. rules).
'
Private Function dstInEffect(ByVal dayNum as UnsignedInteger, ByVal seconds as UnsignedLong) As Boolean
	Const DaysPerWeek as Byte = 7
	Const dstChangeSeconds as UnsignedLong = 2 * SecondsPerHour

	dstInEffect = False

	Dim year as Integer
	Dim month as Byte
	Dim day as Byte

	' determine the month and day corresponding to the day number
	Call GetDate(year, month, day, dayNum)
	If ((month > 3) And (month < 11)) Then
		dstInEffect = True
	ElseIf ((month = 3) Or (month = 11)) Then
		Dim mDay as Byte

		' determine the month day of the first Sunday in the month
		mDay = CByte(DaysPerWeek - GetDayOfWeek(dayNum) + 1)

		If (month = 3) Then
			' determine the month day of the second Sunday
			mDay = mDay + 7
			If ((day > mDay) Or ((day = mDay) And (seconds >= dstChangeSeconds))) Then
				dstInEffect = True
			End If
		ElseIf (month = 11) Then
			' determine the month day of the last Sunday
			If (mDay < 3) Then
				mDay = mDay + 28	' the fifth Sunday
			Else
				mDay = mDay + 21	' the fourth Sunday
			End If
			If ((day < mDay) Or ((day = mDay) And (seconds < dstChangeSeconds))) Then
				dstInEffect = True
			End If
		End If
	End If
End Function

'------------------------------------------------------------------------------
'
'' dspIpAddr
'
' Output an IP address in dotted quad notation (e.g. 192.168.0.45).
'
Private Sub dspIpAddr(ByRef buf() as Byte, ByVal newline as Boolean = False)
	Dim i as Byte
	For i = 1 to 4
		If (i > 1) Then
			Debug.Print ".";
		End If
		Debug.Print buf(i);
	Next i
	If (newline) Then
		Debug.Print
	End If
End Sub

'------------------------------------------------------------------------------
'
'' dspIpAddr
'
' Output an IP address in dotted quad notation (e.g. 192.168.0.45).
'
Private Sub dspIpAddr(ByRef ip as IPAddress, ByVal newline as Boolean = False)
	Dim addr as UnsignedInteger
	addr = ip.addr()
	Dim buf() as Byte Based addr
	Call dspIpAddr(buf, newline)
End Sub

'------------------------------------------------------------------------------
'
'' outputDigits
'
' Output two digits corresponding to a passed value with an optional prefix.
'
Private Sub outputDigits(ByVal val as Byte, ByVal prefix as String = "")
	Dim digit1 as Byte
	Dim digit2 as Byte
	digit2 = (val Mod 10)
	val = val \ 10
	digit1 = (val Mod 10)
	Debug.Print prefix; digit1; digit2;
End Sub

'----------------------------------------------------------------------------
'
'' StrPtr
'
' This helper function is used to convert the address where a string's characters
' are stored into a pointer-to-a-byte as needed by some of the Arduino network
' functions.  Note that Arduino network functions expect a string to be in RAM -
' a circumstance that may or may not be true for any particular string in ZBasic.
' You can force a string to reside in RAM by performing an operation on it like
' Left(), Right(), Mid(), etc.  Alternatively, the NullTermString() function below
' may be used to ensure (a) that the string resides in RAM and (b) that it is
' null-terminated as C/C++ functions expect.
'
Function StrPtr(ByRef st as String) as Byte ByRef
	StrPtr.DataAddress = StrAddress(st)
End Function

'----------------------------------------------------------------------------
'
'' NullTermString
'
' Given a string (of any length less than the maximum string length)
' and located in RAM, Program Memory or Persistent Memory, return a
' new string that is guaranteed to reside in RAM and is null terminated.
'
Function NullTermString(ByVal st as String) as String
    Const MaxStrLen as Integer = 255
    Dim slen as Integer
    NullTermString = ""
    slen = Len(st)
    If ((slen = 0) Or (Asc(st, slen) <> 0)) Then
        ' zero-length string or string not already null terminated
        If (slen < 2) Then
            ' resulting string will be 1 or 2 characters, construct in-situ
            Dim strData() as Byte Based NullTermString.DataAddress
            strData(1) = CByte(slen + 1)
            strData(2) = &He5
            strData(3) = IIF(slen = 0, 0, Asc(st, 1))
            strData(4) = 0
        ElseIf (slen < MaxStrLen) Then
            ' resulting string will be at least 3 characters in allocated RAM
            NullTermString = st & Chr(0)
        End If
    Else
        ' string is already null-terminated, ensure it is in RAM
        Select Case StrType(st)
        Case &He2, &He3
            ' source string is in ProgMem or Persistent Memory
            NullTermString = Left(st, slen)
        Case Else
            ' source string is already in RAM
            NullTermString = st
        End Select
    End If
End Function

'----------------------------------------------------------------------------
'
'' getHostAddress
'
' Given a host name, attempt to resolve it to an IP address.  Return True if
' successful, False otherwise.
'
Private Function getHostAddress(ByRef eth as EthernetBasicClass, ByVal hostName as String, _
		ByRef hostIP as IPAddress) as Boolean
	getHostAddress = False
#if defined(NET_TIME_DNS)
	Dim dns as DNSClient
	Dim st as String
	st = NullTermString(hostName)
	Call dns.begin(eth.dnsServerIP())
	getHostAddress = (dns.getHostByName(StrPtr(st), hostIP) = 1)
#endif
End Function

#if defined(INCLUDE_TEST_DRIVER)

  #if Not defined(NET_TIME_DHCP)
Dim Ethernet as EthernetBasicClass
  #endif

'-------------------------------------------------------------------------------
'
'' Main
'
' This simple driver illustrates how to use the GetTimeNTP() function to set the
' ZBasic RTC to the current time.
'
Sub Main()
	Dim buf(1 to 10) as Byte	' buffer to hold incoming and outgoing packets 

	' initialize the SPI channel to use (zbChanSPI defined in SPI/SPI.h)
	Call OpenSPI(zbChanSPI, &H80, csPin)

	' initialize the Ethernet object, get an IP address
	Call GetProgMem(@myMacAddr, buf, SizeOf(myMacAddr))
  #if defined(NET_TIME_DHCP)
	Debug.Print "Acquiring IP address...."
	If (Ethernet.begin(buf) = 0) Then
		Debug.Print "Failed to get IP address via DHCP"
	Else
		Call W5100.getIPAddress(buf)
		Debug.Print "IP address: ";
		Call dspIpAddr(buf, True)
	End If
  #else
    #if defined(NET_TIME_DNS)
	Call Ethernet.begin(buf, ipAddr, dnsAddr, gwAddr)
	#else
	Call Ethernet.begin(buf, ipAddr, INADDR_NONE, gwAddr)
	#endif
	Call W5100.getIPAddress(buf)
	Debug.Print "IP address: ";
	Call dspIpAddr(buf, True)
  #endif
	
	Dim seconds as UnsignedLong
	seconds =  GetTimeNTP(Ethernet)
	If (seconds > 0) Then
		' determine the day number (relative to 1/1/1999)
		Dim dayNum as UnsignedInteger
		seconds = seconds - DeltaZBasicNTP
		dayNum = CUInt(seconds \ SecondsPerDay)
		seconds = seconds Mod SecondsPerDay

		' set the ZBasic RTC
		Atomic
			Register.RTCDay = dayNum
			Register.RTCTick = CLng(seconds * CULng(Register.RTCTickFrequency))
		End Atomic

		' display the time
		Dim year as Integer
		Dim month as Byte
		Dim day as Byte
		Dim hour as Byte
		Dim minute as Byte
		Dim sec as Single
		Call GetTimestamp(year, month, day, hour, minute, sec)
		Debug.Print "time set to";
		Call outputDigits(day, " ")
		Debug.Print " "; monthName(month); " "; year;
		Call outputDigits(hour, " ")
		Call outputDigits(minute, ":")
		Call outputDigits(CByte(sec), ":")
		Debug.Print
	Else
		Debug.Print "no reply from NTP server"
	End If
	
	Do
  #if defined(NET_TIME_DHCP)
		' maintain DHCP lease
		Call Ethernet.maintain()
  #endif
	Loop
End Sub
#endif

