'------------------------------------------------------------------------------
' 
' 	BSD License
'   -----------
' 
'   Copyright (c) 2013 Elba Corp.  All rights reserved.
' 
'      Elba Corp.
'      10915 SW Avocet Ct.
'      Beaverton, OR  97007
' 
'   Redistribution and use in source and binary forms, with or without
'   modification, are permitted only for use by licensees of the ZBasic
'   Compiler for generic AVR devices and only then if all the following
'   conditions are met:
' 
'   - Redistributions of source code must retain the above copyright notice,
'     this list of conditions and the following disclaimer.
' 
'   - Redistributions in binary form must reproduce the above copyright notice,
'     this list of conditions and the following disclaimer in the documentation
'     and/or other materials provided with the distribution.
' 
'   - Neither the name of the copyright owner nor the names of its contributors
'     may be used to endorse or promote products derived from this software
'     without specific prior written permission of the copyright owner.
' 
'   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
'   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
'   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
'   ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
'   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
'   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
'   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
'   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
'   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
'   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
'   POSSIBILITY OF SUCH DAMAGE.
' 
'------------------------------------------------------------------------------

'------------------------------------------------------------------------------
'
' This module demonstrates how to use the Arduino Ethernet classes to implement
' a simple UDP server.  The server accepts two commands:
'
'    W<pin>,<val>
'    R<pin>
'
' where <pin> may be a physical pin designator (e.g. 12) or a port/pin designator
' (e.g. C.0) and <val> is either 0 or 1.  If the specified pin is valid for the
' target device the W command causes the designated pin to be set to the specified
' state and the response "ack" is returned.  For the R command, the current state
' of the designated pin is returned, e.g. "0" or "1".  For badly formed or
' unrecognized commands, the response "nak" is returned.  This simple command
' structure can be expanded to meet a variety of needs.
'
' For testing purposes, an application for sending UDP commands will be needed.
' One example of such an application is the command line app UDPSZ by Luigi
' Auriemma (http://aluigi.altervista.org/testz/udpsz.zip).
'
' There are several configurable elements (unless otherwise indicated, found
' in this file):
'
'    - the ZBasic SPI channel to use (specified in in Arduino/SPI/SPI.h)
'    - the SPI chip select pin of the W5100 chip
'    - the MAC address for the W5100 chip
'    - the IP address for the server and the gateway IP address (if not using DHCP)
'    - the port on which to listen for UDP requests
'
' Although it shouldn't need to be changed, the client port address to use
' can also be modified.
' 
'------------------------------------------------------------------------------

' allow Functions to be invoked as if they are Subroutines
Option CallFunctions

' specify Arduino compatibility
Option Arduino "C:/Projects/Arduino"

' import the necessary Arduino Ethernet declarations
#import "UdpServerImport.h" "#UDP_SERVER_DHCP"

' specify the CS pin for the W5100 chip
Private Const csPin as Byte = 14

' specify the MAC address of the W5100 chip (must be unique on a local network)
Private myMacAddr as ByteVectorData ({ &H00, &H10, &H5a, &H63, &H2e, &H85 })

#if Not defined(UDP_SERVER_DHCP)
	' when DHCP is not used, specify the host and gateway IP addresses
	Private ipAddr as IPAddress = _Create(192, 168, 0, 31)
	Private gwAddr as IPAddress = _Create(192, 168, 0, 2)
#endif

' specify the port to which you want the UDP server to listen for connections
Private Const localPort as UnsignedInteger = 8888

' uncomment the line below to output information about UDP requests received
#define DEBUG_UDP_REQUEST

#if Not defined(UDP_SERVER_DHCP)
#endif

'------------------------------------------------------------------------------
'
'' Main
'
' This example code implements a simple UDP server.
'
Sub Main()
	Dim buf(1 to 100) as Byte

	' initialize the SPI channel to use (spiChan defined in SPI/SPI.h)
	Call OpenSPI(zbChanSPI, &H80, csPin)

	' get the adapter MAC address into the buffer
	Call GetProgMem(@myMacAddr, buf, SizeOf(myMacAddr))

#if defined(UDP_SERVER_DHCP)
	' Initialize the Ethernet object, get an IP address via DHCP
	' (note that the EthernetClass object is defined in Arduino code).
	Debug.Print "Acquiring IP address...."
	If (Ethernet.begin(buf) = 0) Then
		Debug.Print "Failed to get IP address via DHCP"
	Else
		Call W5100.getIPAddress(buf)
		Debug.Print "IP address: ";
		Call dspIpAddr(buf, True)
	End If

	' define the Udp object
	Dim Udp as EthernetUDP
#else
	' no DHCP, initialize the Ethernet object
	Dim Ethernet as EthernetBasicClass
	Call Ethernet.begin(buf, ipAddr, INADDR_NONE, gwAddr)
	Call W5100.getIPAddress(buf)
	Debug.Print "IP address: ";
	Call dspIpAddr(buf, True)

	' define the Udp object
	Dim Udp as EthernetBasicUDP
#endif

	' initialize the UDP object
	Call Udp.begin(localPort)
	Do
#if defined(UDP_SERVER_DHCP)
		' maintain DHCP lease
		Call Ethernet.maintain()
#endif

		Dim pktSize as Integer
		pktSize = Udp.parsePacket()
		If (pktSize > 0) Then
			' retrieve the client IP address and port for later use
			Dim remoteIP as IPAddress
			Dim remotePort as UnsignedInteger
			remoteIP = Udp.remoteIP()
			remotePort = Udp.remotePort()

#if defined(DEBUG_UDP_REQUEST)
			' report the receipt of a UDP packet
			Debug.Print "received UDP packet ("; pktSize; " bytes) from ";
			Call getIpAddr(buf, remoteIP)
			Call dspIpAddr(buf)
			Debug.Print ":"; remotePort
#endif
			' limit the packet size to the size of the buffer
			Dim readCount as Integer
			readCount = Min(pktSize, (SizeOf(buf) - 1))

			' read the packet into a buffer, read and discard the excess
			Call Udp.read(buf, CUInt(readCount))
			buf(readCount + 1) = 0
			Do While (readCount < pktSize)
				Call Udp.read()
				pktSize = pktSize - 1
			Loop

#if defined(DEBUG_UDP_REQUEST)
			' output the packet content
			Debug.Print "Packet contents (in hex):"
			Dim i as Integer, cnt as Integer
			cnt = 0
			For i = 1 to pktSize
				If (cnt > 0) Then
					Debug.Print " ";
				End If
				Debug.Print CStrHex(buf(i));
				cnt = cnt + 1
				If (cnt >= 16) Then
					Debug.Print
					cnt = 0
				End If
			Next i
			If (cnt > 0) Then
				Debug.Print
			End If
#endif

			' parse the command in the packet
			Dim idx as Integer
			Dim reply(1 to 4) as Byte
			Dim pin as Byte
			idx = 2
			reply(1) = 0
			Select Case buf(1)
			Case Asc("W")	' write a value to a pin
				' read the pin designator
				pin = getPinNumber(buf, idx)
				If ((pin <> 0) And (buf(idx) = Asc(","))) Then
					' have valid pin designator followed by comma
					idx = idx + 1

					' validate the desired pin state
					If (((buf(idx) = Asc("0")) Or (buf(idx) = Asc("1"))) And (buf(idx + 1) = 0)) Then
						' set the pin state, indicate success
						Call PutPin(pin, buf(idx) - Asc("0"))
						reply(1) = Asc("a")
						reply(2) = Asc("c")
						reply(3) = Asc("k")
						reply(4) = 0
					End If
				End If

			Case Asc("R")	' read a pin
				' read the pin designator
				pin = getPinNumber(buf, idx)
				If ((pin <> 0) And (buf(idx) = 0)) Then
					' set the reply to indicate the state of the pin
					reply(1) = GetPin(pin) + Asc("0")
					reply(2) = 0
				End If

			End Select

			' if no reply has been set, the command is malformed or unrecognized
			If (reply(1) = 0) Then
				reply(1) = Asc("n")
				reply(2) = Asc("a")
				reply(3) = Asc("k")
				reply(4) = 0
			End If

			' send a reply, to the IP address and port that sent the packet we received
			Call Udp.beginPacket(remoteIP, remotePort)
			Call Udp.write(reply)
			Call Udp.endPacket()
		End If
	Loop
End Sub

'------------------------------------------------------------------------------
'
'' getPinNumber
'
' Extract a pin designator from a buffer beginning at the given index, returning
' the equivalent pin number or zero if unrecognized or invalid for the target
' device.  A pin number may be specified using the notation P.n (where P is a
' letter specifying the port and n is a digit from 0-7 indicating the bit number
' of the port) or by using the physical pin number.
'
Private Function getPinNumber(ByRef buf() as Byte, ByRef idx as Integer) as Byte
	Dim b as Byte
	getPinNumber = 0
	b = toUpperChar(buf(idx))
	If (isCharDigit(b)) Then
		' physical pin number, get the decimal equivalent
		Do
			getPinNumber = (getPinNumber * 10) + (b - Asc("0"))
			idx = idx + 1
			b = buf(idx)
		Loop While (isCharDigit(b))

		' validate the pin number, convert to port/bit form
		getPinNumber = PortBit(getPinNumber)

	ElseIf (isCharAlpha(b) And (buf(idx + 1) = Asc("."))) Then
		' possible port.bit form
		Dim portIdx as Byte
		portIdx = b - Asc("A")
		idx = idx + 2
		b = buf(idx)
		If ((b >= Asc("0")) And (b <= Asc("7"))) Then
			' validate the port/bit combination
			getPinNumber = PortBit(portIdx, b - Asc("0"))
			idx = idx + 1
		End If
	End If
End Function

'------------------------------------------------------------------------------
'
'' isCharDigit
'
' Determine if a character value represents a decimal digit.
'
Private Function isCharDigit(ByVal c as Byte) as Boolean
	isCharDigit = ((c >= Asc("0")) And (c <= Asc("9")))
End Function

'------------------------------------------------------------------------------
'
'' isCharAlpha
'
' Determine if a character value represents an alphabetic character (A-Z or a-z).
'
Private Function isCharAlpha(ByVal c as Byte) as Boolean
	isCharAlpha = (((c >= Asc("A")) And (c <= Asc("Z"))) Or _
			((c >= Asc("a")) And (c <= Asc("z"))))
End Function

'------------------------------------------------------------------------------
'
'' toUpperChar
'
' Convert a character value to upper case if it is lower case.
'
Private Function toUpperChar(ByVal c as Byte) as Byte
	If ((c >= Asc("a")) And (c <= Asc("z"))) Then
		toUpperChar = c - Asc("a") + Asc("A")
	Else
		toUpperChar = c
	End If
End Function

'------------------------------------------------------------------------------
'
'' dspIpAddr
'
' Output an IP address in dotted quad notation (e.g. 192.168.1.25).
'
Private Sub dspIpAddr(ByRef buf() as Byte, ByVal newline as Boolean = False)
	Dim i as Byte
	For i = 1 to 4
		If (i > 1) Then
			Debug.Print ".";
		End If
		Debug.Print buf(i);
	Next i
	If (newline) Then
		Debug.Print
	End If
End Sub

'------------------------------------------------------------------------------
'
'' getIpAddr
'
' Copy an IP address from an IPAddress object to a buffer.  This helper
' routine is necessary because ZBasic cannot directly use the indexing
' operator overload method provided by the C++ class IPAddress.
'
Private Sub getIpAddr(ByRef buf() as Byte, ByRef ip as IPAddress)
#c
	for (int i = 0; i < 4; i++)
		zp_buf[i] = (*zp_ip)[i];
#endc
End Sub

