'------------------------------------------------------------------------------
' 
' 	BSD License
'   -----------
' 
'   Copyright (c) 2013 Elba Corp.  All rights reserved.
' 
'      Elba Corp.
'      10915 SW Avocet Ct.
'      Beaverton, OR  97007
' 
'   Redistribution and use in source and binary forms, with or without
'   modification, are permitted only for use by licensees of the ZBasic
'   Compiler for generic AVR devices and only then if all the following
'   conditions are met:
' 
'   - Redistributions of source code must retain the above copyright notice,
'     this list of conditions and the following disclaimer.
' 
'   - Redistributions in binary form must reproduce the above copyright notice,
'     this list of conditions and the following disclaimer in the documentation
'     and/or other materials provided with the distribution.
' 
'   - Neither the name of the copyright owner nor the names of its contributors
'     may be used to endorse or promote products derived from this software
'     without specific prior written permission of the copyright owner.
' 
'   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
'   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
'   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
'   ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
'   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
'   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
'   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
'   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
'   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
'   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
'   POSSIBILITY OF SUCH DAMAGE.
' 
'------------------------------------------------------------------------------

'------------------------------------------------------------------------------
'
' This module demonstrates how to use the Arduino Ethernet classes to request
' an HTML page from a web server.  The example uses a specially constructed web
' page on the ZBasic server that returns the current time (in UTC) in the form
' exemplified by the string "time(2012-07-13 15:02:44 1342191764) ".  The
' application then sets the ZBasic RTC using the received data, adjusted for
' local time (controlled by the constants tzAdjMinutes and useDST).
'
' There are several configurable elements (unless otherwise indicated, found
' in this file):
'
'    - the ZBasic SPI channel to use (specified in in Arduino/SPI/SPI.h)
'    - the SPI chip select pin of the W5100 chip
'    - the MAC address for the W5100 chip
'    - the server name or IP address (if not using DHCP)
'    - the URL of the web server page to load
'    - the offset from GMT to the local time zone
'    - if daylight savings time should be observed
'
' Although it shouldn't need to be changed, the client port address to use
' can also be modified.
' 
'------------------------------------------------------------------------------

' allow Functions to be invoked as if they are Subroutines
Option CallFunctions

' specify Arduino compatibility
Option Arduino "C:/Projects/Arduino"

' import the necessary Arduino Ethernet declarations
#import "WebClientImport.h" "#WEB_CLIENT_DNS", "#WEB_CLIENT_DHCP"

' specify the CS pin for the W5100 chip
Private Const csPin as Byte = 14

' specify the MAC address of the W5100 chip (must be unique on a local network)
Private myMacAddr as ByteVectorData ({ &H00, &H10, &H5a, &H63, &H2e, &H85 })

' specify the IP address of the web server
#if defined(WEB_CLIENT_DNS)
	Private Const serverName as String = "www.zbasic.net"
#else
	Private serverIP as IPAddress = _Create(69, 89, 26, 47)	' zbasic.net server
#endif

#if Not defined(WEB_CLIENT_DHCP)
	' specify the host IP address and gateway IP address
	Dim ipAddr as IPAddress = _Create(192, 168, 0, 31)
	Dim gwAddr as IPAddress = _Create(192, 168, 0, 2)
  #if Defined(WEB_CLIENT_DNS)
	' specify the DNS server IP address
	Dim dnsAddr as IPAddress =  _Create(192, 168, 0, 2)
  #endif
#endif

' specify the web page to load
Private Const requestStr as String = "GET /EtherTest.php HTTP/1.1" & Chr(&H0a)

' specify the local time zone relative to GMT
Private Const tzAdjMinutes as Integer = CInt(-8 * MinutesPerHour)

' indicate if daylight savings time algorithm should be used - see dstInEffect()
Private Const useDST as Boolean = True

' specify the standard TCP port number to use
Private Const httpPort as UnsignedInteger = 80		' standard HTTP port

' uncomment one or more of the lines below to enable debug output
'#define DEBUG_CONNECTION
'#define DEBUG_HTTP_SEND
'#define DEBUG_HTTP_RECV

'-------------------------------------------------------------------------------
' conversion constants
Private Const tzAdjustSeconds as UnsignedLong = CLng(tzAdjMinutes * CInt(SecondsPerMinute))
Private Const SecondsPerMinute as UnsignedLong = 60
Private Const MinutesPerHour as UnsignedLong = 60
Private Const HoursPerDay as UnsignedLong = 24
Private Const SecondsPerHour as UnsignedLong = MinutesPerHour * SecondsPerMinute
Private Const SecondsPerDay as UnsignedLong = HoursPerDay * SecondsPerHour
Private Const DeltaZBasicUnix as UnsignedLong = 915148800	' seconds btw 1970/1/1 and 1999/1/1

' This Program Memory array gives the three-letter abbreviation
' for months numbered 1 to 12.
Private monthName as StringVectorData ({
	"Jan", "Feb", "Mar", "Apr", "May", "Jun"
	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
})

#if defined(WEB_CLIENT_DHCP)
	Dim myClient as EthernetClient
#else
	' use the basic classes
	Dim Ethernet as EthernetBasicClass
	Dim myClient as EthernetBasicClient
#endif

'-------------------------------------------------------------------------------
'
'' Main
'
' This simple driver illustrates how to request content from a web server and
' extract information from that content.  In this case, a special page on the
' ZBasic web server returns content like the following (ignoring the leading
' apostrophe and space):
'
' HTTP/1.1 200 OK
' Date: Mon, 16 Jul 2012 17:06:28 GMT
' Server: Apache
' Vary: Accept-Encoding
' Transfer-Encoding: chunked
' Content-Type: text/html
' 
' 45
' <html><body>
' time(1342458389 2012-07-16 17:06:29)<br>
' </body></html>
' 
' 0
'
' The PHP for the web page looks like this (again ignoring the leading
' apostrophe and space):
'
' <html><body>
' <?php
' echo "time(". gmdate("U Y-m-d H:i:s") . ")<br>\n";
' ?>
' </body></html>
'
' The code extracts value of the Unix-format UTC-based timestamp and uses
' it to set the ZBasic RTC to the current time.
'
Sub Main()
	Dim buf(1 to 80) as Byte	' multi-purpose buffer

	' Initialize the SPI channel to use (spiChan defined in SPI/SPI.h)
	Call OpenSPI(zbChanSPI, &H80, csPin)

	' initialize the Ethernet object, get an IP address
	Call GetProgMem(@myMacAddr, buf, SizeOf(myMacAddr))
#if defined(WEB_CLIENT_DHCP)
	' using DHCP, get the host IP address
	Debug.Print "Acquiring IP address...."
	If (Ethernet.begin(buf) = 0) Then
		Debug.Print "Failed to get IP address via DHCP"
	Else
		Call W5100.getIPAddress(buf)
		Debug.Print "IP address: ";
		Call dspIpAddr(buf, True)
	End If
#else
  #if defined(WEB_CLIENT_DNS)
	Call Ethernet.begin(buf, ipAddr, dnsAddr, gwAddr)
  #else
	Call Ethernet.begin(buf, ipAddr, INADDR_NONE, gwAddr)
  #endif
	Call W5100.getIPAddress(buf)
	Debug.Print "IP address: ";
	Call dspIpAddr(buf, True)
#endif
	
	' attempt to connect to the server
	Dim stat as Integer
#if defined(WEB_CLIENT_DNS)
	Dim serverIP as IPAddress
	If (Not getHostAddress(Ethernet, serverName, serverIP)) Then
		Debug.Print "failed to resolve host name """; serverName; """";
		Exit Sub
	End If
  #if 0
	Debug.Print "resolved """; serverName; """ to ";
	Call dspIpAddr(serverIP, True)
  #endif
#endif
	stat = myClient.connect(serverIP, httpPort)

	If (stat <> 0) Then
		Dim lastChar as Integer
		lastChar = 0
#if defined(DEBUG_CONNECTION)
		Debug.Print "connected to server"
#endif

		' send the request string
		Call sendClient(myClient, requestStr)

		' send the host header
		Call sendClient(myClient, "Host: ")
#if defined(WEB_CLIENT_DNS)
		' emit the server name
		Call sendClient(myClient, serverName)
#else
		' emit the IP address
		Dim j as Integer
		Call getIpAddr(buf, serverIp)
		For j = 1 to 4
			If (j > 1) Then Call sendClient(myClient, ".")
			Call sendClient(myClient, CStr(buf(j)))
		Next j
#endif
		Call sendClient(myClient, ":")
		Call sendClient(myCLient, CStr(httpPort))
		Call sendClient(myClient, Chr(&H0a))

		' finish the request with a blank line
		Call sendClient(myClient, Chr(&H0a))

		Dim idx as Integer
		idx = 1
		Dim content as Boolean
		content = False
		Do
			' see if there is more data from the server
			If (myClient.available() > 0) Then
				Dim c as Integer
				c = myClient.read()
				lastChar = c
				If ((c <> &H0d) And (c <> &H0a) And (idx < (UBound(buf) - 1))) Then
					buf(idx) = CByte(c)
					idx = idx + 1
				End If
#if defined(DEBUG_HTTP_RECV)
				If (c = &H0a) Then
					Debug.Print
				ElseIf (c <> &H0d) Then
					Debug.Print Chr(c);
				End If
#endif

				' check for EOL
				If (c = &H0a) Then
					buf(idx) = 0		' null terminate the buffer

					' process content
					If (content) Then
						Dim bufIdx as Integer
						bufIdx = 1
						If (strMatch(buf, bufIdx, "time(")) Then
							If (setTime(getValue(buf, bufIdx))) Then
								Dim year as Integer
								Dim month as Byte
								Dim day as Byte
								Dim hour as Byte
								Dim minute as Byte
								Dim sec as Single
								Call GetTimestamp(year, month, day, hour, minute, sec)
								Debug.Print "time set to";
								Call outputDigits(day, " ")
								Debug.Print " "; monthName(month); " "; year;
								Call outputDigits(hour, " ")
								Call outputDigits(minute, ":")
								Call outputDigits(CByte(sec), ":")
								Debug.Print
							End If
						End If
					End If

					' detect the blank line separating header and content
					If (idx = 1) Then content = True
					idx = 1
				End If
			End If

			' check for a disconnect
			If Not CBool(myClient.connected()) Then
#if defined(DEBUG_CONNECTION)
				If (lastChar <> &H0a) Then Debug.Print
				Debug.Print "disconnecting"
#endif
				Call myClient.stop()
				Exit Do
			End If
		Loop
#if defined(DEBUG_CONNECTION)
	Else
		Debug.Print "failed to connect to server"
#endif
	End If
End Sub

'------------------------------------------------------------------------------
'
'' strMatch
'
' This function determines if the specified string exists in the buffer beginning
' at the specified index.  If so, the index is advanced beyond the string and
' True is returned.
'
Private Function strMatch(ByRef buf() as Byte, ByRef idx as Integer, ByVal s as String) as Boolean
	strMatch = False
	Dim slen as Integer
	slen = Len(s)
	If (slen > 0) Then
		Dim i as Integer
		strMatch = True
		For i = 1 to slen
			If (buf(idx + i - 1) <> Asc(s, i)) Then
				strMatch = False
				Exit For
			End If
		Next i
		If (strMatch) Then
			idx = idx + slen
		End If
	End If
End Function

'------------------------------------------------------------------------------
'
'' getValue
'
' This function attempts to extract a decimal value from the characters in a
' buffer beginning at the specified index.  The resulting value is returned
' directly and the index of the first non-digit character encountered is
' returned indirectly via the 'idx' parameter.
'
Private Function getValue(ByRef buf() as Byte, ByRef idx as Integer) as UnsignedLong
	Dim b as Byte
	getValue = 0
	Do While (isCharDigit(buf(idx)))
		getValue = (getValue * 10) + CULng(buf(idx) - Asc("0"))
		idx = idx + 1
	Loop
End Function

'------------------------------------------------------------------------------
'
'' isCharDigit
'
' Determine if a character value represents a decimal digit.
'
Private Function isCharDigit(ByVal c as Byte) as Boolean
	isCharDigit = ((c >= Asc("0")) And (c <= Asc("9")))
End Function

'------------------------------------------------------------------------------
'
'' setTime
'
' This function sets the ZBasic RTC using the passed value representing the
' number of seconds since 01 Jan 1970 00:00:00 UTC (i.e. a Unix timestamp).
' The time is adjusted for the local time zone and daylight savings time
' (if applicable).
'
' The return value indicates whether the time was successfully set.
'
Private Function setTime(ByVal seconds as UnsignedLong) as Boolean
	setTime = (seconds >= DeltaZBasicUnix)
	If (setTime) Then
		' adjust the seconds count for the ZBasic epoch (01 Jan 1999) and timezone
		seconds = seconds - DeltaZBasicUnix + tzAdjustSeconds

		' determine the day number
		Dim dayNum as UnsignedInteger
		dayNum = CUInt(seconds \ SecondsPerDay)
		seconds = seconds Mod SecondsPerDay

		' adjust the time for DST, if applicable
		If (useDST And dstInEffect(dayNum, seconds)) Then
			seconds = seconds + SecondsPerHour
			If (seconds >= SecondsPerDay) Then
				seconds = seconds - SecondsPerDay
				dayNum =dayNum + 1
			End If
		End If

		' set the ZBasic RTC
		Atomic
			Register.RTCDay = dayNum
			Register.RTCTick = CLng(seconds * CULng(Register.RTCTickFrequency))
		End Atomic
	End If
End Function

'------------------------------------------------------------------------------
'
'' FlashString
'
' This helper function converts a Program Memory address to a reference to
' a __FlashStringHelper used in the Arduino code.
'
Private Function FlashString(ByVal progMemAddr as Long) as __FlashStringHelper ByRef
	FlashString.DataAddress = progMemAddr
End Function

'------------------------------------------------------------------------------
'
'' sendClient
'
' Send the contents of a ZBasic string to the Ethernet client.  Because the
' string's characters could be in RAM, Program Memory or Persistent Memory, the
' characters are extracted one-by-one and sent to the client individually.
'
Private Sub sendClient(ByRef theClient as EthernetBasicClient, ByVal s as String)
	Dim i as Integer
	For i = 1 to Len(s)
		Dim c as Byte
		c = Asc(s, i)
		Call theClient.write(c)
#if defined(DEBUG_HTTP_SEND)
		If (c = LF) Then
			Debug.Print
		ElseIf (c <> CRF) Then
			Debug.Print Chr(c);
		End If
#endif
	Next i
End Sub

'-------------------------------------------------------------------------------
'
'' dstInEffect
'
' Determine if Daylight Saving Time is in effect as of the ZBasic day number and
' second count specified.  The start/stop dates are the second Sunday in March 
' and the first Sunday of November (i.e. the 2007 U.S. rules).
'
Private Function dstInEffect(ByVal dayNum as UnsignedInteger, ByVal seconds as UnsignedLong) As Boolean
	Const DaysPerWeek as Byte = 7
	Const dstChangeSeconds as UnsignedLong = 2 * SecondsPerHour

	dstInEffect = False

	Dim year as Integer
	Dim month as Byte
	Dim day as Byte

	' determine the month and day corresponding to the day number
	Call GetDate(year, month, day, dayNum)
	If ((month > 3) And (month < 11)) Then
		dstInEffect = True
	ElseIf ((month = 3) Or (month = 11)) Then
		Dim mDay as Byte

		' determine the month day of the first Sunday in the month
		mDay = CByte(DaysPerWeek - GetDayOfWeek(dayNum) + 1)

		If (month = 3) Then
			' determine the month day of the second Sunday
			mDay = mDay + 7
			If ((day > mDay) Or ((day = mDay) And (seconds >= dstChangeSeconds))) Then
				dstInEffect = True
			End If
		ElseIf (month = 11) Then
			' determine the month day of the last Sunday
			If (mDay < 3) Then
				mDay = mDay + 28	' the fifth Sunday
			Else
				mDay = mDay + 21	' the fourth Sunday
			End If
			If ((day < mDay) Or ((day = mDay) And (seconds < dstChangeSeconds))) Then
				dstInEffect = True
			End If
		End If
	End If
End Function

'------------------------------------------------------------------------------
'
'' dspIpAddr
'
' Output an IP address in dotted quad notation (e.g. 192.168.0.45).
'
Private Sub dspIpAddr(ByRef buf() as Byte, ByVal newline as Boolean = False)
	Dim i as Byte
	For i = 1 to 4
		If (i > 1) Then
			Debug.Print ".";
		End If
		Debug.Print buf(i);
	Next i
	If (newline) Then
		Debug.Print
	End If
End Sub

'------------------------------------------------------------------------------
'
'' dspIpAddr
'
' Output an IP address in dotted quad notation (e.g. 192.168.0.45).
'
Private Sub dspIpAddr(ByRef ip as IPAddress, ByVal newline as Boolean = False)
	Dim addr as UnsignedInteger
	addr = ip.addr()
	Dim buf() as Byte Based addr
	Call dspIpAddr(buf, newline)
End Sub

'------------------------------------------------------------------------------
'
'' outputDigits
'
' Output two digits corresponding to a passed value with an optional prefix.
'
Private Sub outputDigits(ByVal val as Byte, ByVal prefix as String = "")
	Dim digit1 as Byte
	Dim digit2 as Byte
	digit2 = (val Mod 10)
	val = val \ 10
	digit1 = (val Mod 10)
	Debug.Print prefix; digit1; digit2;
End Sub

'----------------------------------------------------------------------------
'
'' StrPtr
'
' This helper function is used to convert the address where a string's characters
' are stored into a pointer-to-a-byte as needed by some of the Arduino network
' functions.  Note that Arduino network functions expect a string to be in RAM -
' a circumstance that may or may not be true for any particular string in ZBasic.
' You can force a string to reside in RAM by performing an operation on it like
' Left(), Right(), Mid(), etc.  Alternatively, the NullTermString() function below
' may be used to (a) ensure that the string resides in RAM and (b) that it is
' null-terminated as C/C++ functions expect.
'
Function StrPtr(ByRef st as String) as Byte ByRef
	StrPtr.DataAddress = StrAddress(st)
End Function

'----------------------------------------------------------------------------
'
'' NullTermString
'
' Given a string (of any length less than the maximum string length)
' and located in RAM, Program Memory or Persistent Memory, return a
' new string that is guaranteed to reside in RAM and is null terminated.
'
Function NullTermString(ByVal st as String) as String
    Const MaxStrLen as Integer = 255
    Dim slen as Integer
    NullTermString = ""
    slen = Len(st)
    If ((slen = 0) Or (Asc(st, slen) <> 0)) Then
        ' zero-length string or string not already null terminated
        If (slen < 2) Then
            ' resulting string will be 1 or 2 characters, construct in-situ
            Dim strData() as Byte Based NullTermString.DataAddress
            strData(1) = CByte(slen + 1)
            strData(2) = &He5
            strData(3) = IIF(slen = 0, 0, Asc(st, 1))
            strData(4) = 0
        ElseIf (slen < MaxStrLen) Then
            ' resulting string will be at least 3 characters in allocated RAM
            NullTermString = st & Chr(0)
        End If
    Else
        ' string is already null-terminated, ensure it is in RAM
        Select Case StrType(st)
        Case &He2, &He3
            ' source string is in ProgMem or Persistent Memory
            NullTermString = Left(st, slen)
        Case Else
            ' source string is already in RAM
            NullTermString = st
        End Select
    End If
End Function

'------------------------------------------------------------------------------
'
'' getIpAddr
'
' Copy an IP address from an IPAddress object to a buffer.  This helper
' routine is necessary because ZBasic cannot directly use the indexing
' operator overload method provided by the C++ class IPAddress.
'
Private Sub getIpAddr(ByRef buf() as Byte, ByRef ip as IPAddress)
#c
	for (int i = 0; i < 4; i++)
		zp_buf[i] = (*zp_ip)[i];
#endc
End Sub

'----------------------------------------------------------------------------
'
'' getHostAddress
'
' Given a host name, attempt to resolve it to an IP address.  Return True if
' successful, False otherwise.
'
Private Function getHostAddress(ByRef eth as EthernetBasicClass, ByVal hostName as String, _
		ByRef hostIP as IPAddress) as Boolean
	getHostAddress = False
#if defined(WEB_CLIENT_DNS)
	Dim dns as DNSClient
	Dim st as String
	st = NullTermString(hostName)
	Call dns.begin(eth.dnsServerIP())
	getHostAddress = (dns.getHostByName(StrPtr(st), hostIP) = 1)
#endif
End Function
