'------------------------------------------------------------------------------
' 
' 	BSD License
'   -----------
' 
'   Copyright (c) 2013 Elba Corp.  All rights reserved.
' 
'      Elba Corp.
'      10915 SW Avocet Ct.
'      Beaverton, OR  97007
' 
'   Redistribution and use in source and binary forms, with or without
'   modification, are permitted only for use by licensees of the ZBasic
'   Compiler for generic AVR devices and only then if all the following
'   conditions are met:
' 
'   - Redistributions of source code must retain the above copyright notice,
'     this list of conditions and the following disclaimer.
' 
'   - Redistributions in binary form must reproduce the above copyright notice,
'     this list of conditions and the following disclaimer in the documentation
'     and/or other materials provided with the distribution.
' 
'   - Neither the name of the copyright owner nor the names of its contributors
'     may be used to endorse or promote products derived from this software
'     without specific prior written permission of the copyright owner.
' 
'   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
'   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
'   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
'   ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
'   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
'   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
'   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
'   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
'   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
'   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
'   POSSIBILITY OF SUCH DAMAGE.
' 
'------------------------------------------------------------------------------

'------------------------------------------------------------------------------
'
' This module implements a simple web server, optionally with dynamic content.
'
' There are several configurable elements (unless otherwise indicated, found
' in this file):
'
'    - the ZBasic SPI channel to use (specified in in Arduino/SPI/SPI.h)
'    - the SPI chip select pin of the W5100 chip
'    - the MAC address for the W5100 chip
'    - the IP address (if not using DHCP)
'    - the port on which to listen for HTTP requests (usually port 80)
'    - the web content to serve
'
' To enable dynamic content, uncomment the definition of QUERY_STRING below
' and edit the code in processQueryStringElem() to recognize the query string
' identifiers and implement the desired action.  The sample code shows how to
' read a pin and set/clear a pin.
'
'------------------------------------------------------------------------------

' Uncomment the line below to enable processing of a query string accompanying
' the web page request.  The query string is appended to the web URL preceded
' by a question mark and contains one or more name/value pairs separated by
' an ampersand.  Example: http://192.168.0.31?getpin=A.3&setpin=A.4
#define QUERY_STRING

' uncomment the line below to output information about HTTP requests received
'#define DEBUG_HTTP_REQUEST

' allow Functions to be invoked as if they are Subroutines
Option CallFunctions

' specify Arduino compatibility
Option Arduino "C:/Projects/Arduino"

' import the necessary Arduino Ethernet declarations
#import "WebServerImport.h" "#WEB_SERVER_DHCP"

' specify the CS pin for the W5100 chip (change to match your configuration)
Private Const csPin as Byte = 14

' specify the MAC address of the W5100 chip (must be unique on a local network)
Private myMacAddr as ByteVectorData ({ &H00, &H10, &H5a, &H63, &H2e, &H85 })

#if Not defined(WEB_SERVER_DHCP)
	' DHCP is not used, specify the host and gateway IP addresses
	Private ipAddr as IPAddress = _Create(192, 168, 0, 31)
	Private gwAddr as IPAddress = _Create(192, 168, 0, 2)
#endif

' specify the port to which you want the web server to listen (Port 80 is the usual HTTP port)
Private Const httpPort as UnsignedInteger = 80

'------------------------------------------------------------------------------

' The next several ProgMem data definitions provide the text for HTTP headers
' for normal and abnormal situations as well as the text for the HTML content
' to be delivered.  The HTML content is split into two parts, the head and the
' tail, to allow for dynamic content to be generated between the two.
'
' header information to send when requested content is unavailable
Dim noContentHeaders as ByteVectorData ({
	"HTTP/1.1 204 No Content", &H0d, &H0a,
	"Connection: close", &H0d, &H0a
	&H0d, &H0a
	&H00
})

' header information to send when requested content is not found
Dim notFoundHeaders as ByteVectorData ({
	"HTTP/1.1 404 Not Found", &H0d, &H0a,
	"Connection: close", &H0d, &H0a
	&H0d, &H0a
	&H00
})

' header information for a successful request
Dim normalHeaders as ByteVectorData ({
	"HTTP/1.1 200 OK", &H0d, &H0a,
	"Content-Type: text/html", &H0d, &H0a,
	"Cache-Control: no-cache", &H0d, &H0a,
	"Connection: close", &H0d, &H0a
	&H0d, &H0a
	&H00
})

' partial webserver content
Dim myHTML_head as ByteVectorData ({
	"<html><body>", &H0d, &H0a,
	"<title>ZBasic Web Server</title>", &H0d, &H0a,
	"<center><h2><font color=""#0000FF"">Wiznet W5100</font></h2>", &H0d, &H0a,
	"This is my ZBasic web server page.<br><br>", &H0d, &H0a,
	&H00
})

Dim myHTML_tail as ByteVectorData ({
	"</center></body></html>", &H0d, &H0a,
	&H00
})

'------------------------------------------------------------------------------

' content types
Private Enum WebContentType
	WebRoot
	WebNoContent
	WebNotFound
End Enum

'------------------------------------------------------------------------------
'
'' Main
'
' This example code implements a simple web server.
'
Sub Main()
	Dim buf(1 to 100) as Byte

	' initialize the SPI channel to use (zbChanSPI defined in SPI/SPI.h)
	Call OpenSPI(zbChanSPI, &H80, csPin)

	' get the adapter MAC address into the buffer
	Call GetProgMem(@myMacAddr, buf, SizeOf(myMacAddr))

#if defined(WEB_SERVER_DHCP)
	' Initialize the Ethernet object, get an IP address via DHCP
	' (note that the EthernetClass object is defined in Arduino code).
	Debug.Print "Acquiring IP address...."
	If (Ethernet.begin(buf) = 0) Then
		Debug.Print "Failed to get IP address via DHCP"
	Else
		Call W5100.getIPAddress(buf)
		Debug.Print "IP address: ";
		Call dspIpAddr(buf, True)
	End If
#else
	' no DHCP, initialize the Ethernet object
	Dim Ethernet as EthernetBasicClass
	Call Ethernet.begin(buf, ipAddr, INADDR_NONE, gwAddr)
	Call W5100.getIPAddress(buf)
	Debug.Print "IP address: ";
	Call dspIpAddr(buf, True)
#endif

	' start listening on the specified port
	Dim myClient as EthernetBasicClient
	Dim myServer as EthernetServer = _Create(httpPort)
	Call myServer.begin()
	Do
#if defined(WEB_SERVER_DHCP)
		' maintain the DHCP lease
		Call Ethernet.maintain()
#endif

		' see if a request has been received
		myClient = myServer.available()
		Dim stat as Byte
		stat = myClient.status()
		If (stat <> SnSR::CLOSED) Then
			Dim content as WebContentType
			Dim queryStr as String
			Dim idx as Integer
			idx = 1
			content = WebNotFound

			' see if a connection exists
			Do While (CBool(myClient.connected()))
				' we have a connection, read the request information and send a reply
				Dim c as Byte
				c = CByte(myClient.read())

#if defined(DEBUG_HTTP_REQUEST)
				' output each character as it is read
				If (c = &H0a) Then
					Debug.Print
				ElseIf (c <> &H0d) Then
					Debug.Print Chr(c);
				End If
#endif
				' check for a full line having been received
				If (c = &H0a) Then
					' check for a a blank line (marking the end of the request header)
					If ((idx = 2) And (buf(1) = &H0d)) Then
						Dim flashAddr as Long
						If (content = WebRoot) Then
							' send the normal headers and partial content
							Call myClient.print(FlashString(@normalHeaders))
							Call myClient.print(FlashString(@myHTML_head))
#if defined(QUERY_STRING)
							Call processQueryString(myClient, queryStr)
#endif

							' arrange to send the remainder of the content
							flashAddr = @myHTML_tail
						ElseIf (content = WebNoContent) Then
							' arrange to send the "no content" headers
							flashAddr = @noContentHeaders
						Else
							' arrange to send the 404 headers
							flashAddr = @notFoundHeaders
						End If
						' send additional information
						Call myClient.print(FlashString(flashAddr))
						Exit Do
					ElseIf ((idx >= 5) And bufMatch(buf, "GET /")) Then
						' detected a GET request, determine what was requested
						Dim reqLen as Integer
						reqLen = idx - 6
						idx = 6
						If ((reqLen >= 1) And ((buf(idx) = Asc(" ")) Or (buf(idx) = Asc("?")))) Then
							' detected a request for the web root
							content = WebRoot
						ElseIf ((reqLen >= 10) And bufMatch(buf(idx), "index.html")) Then
							' detected a request for /index.html
							content = WebRoot
							idx = idx + 10
							reqLen = reqLen - 10
						ElseIf ((reqLen >= 12) And bufMatch(buf(idx), "favicon.ico ")) Then
							' detected a request for the favicon file
							content = WebNoContent
							reqLen = 0
						End If

						' extract the query string
						If ((reqLen > 1) And (buf(idx) = Asc("?"))) Then
							queryStr = MakeString(@buf(idx + 1), reqLen - 1)
						Else
							queryStr = ""
						End If
					End If

					' reset the buffer index for the next request
					idx = 1
				ElseIf (idx < (SizeOf(buf) - 1)) Then
					' store the received character (limited to the size of the buffer less one)
					buf(idx) = c
					idx = idx + 1
				End If
			Loop

			' close the socket
			Call myClient.stop()
		End If
		Call Sleep(0.2)
	Loop
End Sub

'------------------------------------------------------------------------------
'
'' processQueryString
'
' Parse the query string into name/value pairs and perform particular functions
' based on the name.  Multiple name/value pairs are separated by an ampersand.
'
Private Sub processQueryString(ByRef theClient as EthernetBasicClient, ByVal qs as String)
#if defined(DEBUG_HTTP_REQUEST)
	If (Len(qs) > 0) Then
		Call sendClient(theClient, "Query string:[")
		Call sendClient(theClient, qs)
		Call sendClient(theClient, "]<br>" & Chr(&H0d) & Chr(&H0a))
	End If
#endif
	Dim idx as Integer
	Dim qsLen as Integer
	qsLen = Len(qs)
	idx = 1
	Do While (idx <= qsLen)
		Call processQueryStringElem(theClient, qs, idx)
	Loop
End Sub

'------------------------------------------------------------------------------
'
'' processQueryStringElem
'
' Parse a name/value pair (e.g. setpin=5) from a query string beginning at the
' specified string index (1-based).  Multiple name/value pairs are separated by
' an ampersand so the current name/value pair is terminated by either the end
' of the string or an ampersand.
'
' The passed string index is updated in preparation for the next call.
'
Private Sub processQueryStringElem(ByRef theClient as EthernetBasicClient, _
		ByVal qs as String, ByRef idx as Integer)
	' determine the length of the next (possibly last) element in the query string
	Dim elemLen as Integer
	Dim startIdx as Integer
	Dim strIdx as Integer
	Dim sLen as Integer

	' determine the length of the next element
	sLen = Len(qs)
	startIdx = idx
	If (idx < sLen) Then
		' see if an ampersand is present
		strIdx = CInt(StrFind(qs, "&", idx))
		If (strIdx > 0) Then
			' not the last element
			elemLen = strIdx - idx
			idx = strIdx + 1
		Else
			' found the last element
			elemLen = sLen - idx + 1
			idx = sLen + 1
		End If
	Else
		elemLen = 0
	End If
	
	If (elemLen > 0) Then
		Dim value as String
		Dim name as String

		' see if the element contains an equal sign
		strIdx = CInt(StrFind(qs, "=", startIdx))
		If ((strIdx > 0) And (strIdx < idx)) Then
			' extract the name and value components
			name = Mid(qs, startIdx, strIdx - startIdx)
			value = Mid(qs, strIdx + 1, elemLen - (strIdx - startIdx + 1))

			' replace hexadecimal escapes in the value string
			value = convertQueryEscapes(value)
		Else
			' no value, extract the name component
			name = Mid(qs, startIdx, elemLen)
			value = ""
		End If

		Dim pin as Byte
		Select Case name
		Case "setpin"	' setpin=<pin number>
			pin = getPinNumber(value)
			If (pin <> 0) Then
				Call PutPin(pin, 1)
				Call sendClient(theClient, "Set pin """)
			Else
				Call sendClient(theClient, "Error: invalid pin """)
			End If
			Call sendClient(theClient, value)
			Call sendClient(theClient, """<br>" & Chr(&H0d) & Chr(&H0a))

		Case "clrpin"	' clrpin=<pin number>
			pin = getPinNumber(value)
			If (pin <> 0) Then
				Call PutPin(pin, 0)
				Call sendClient(theClient, "Cleared pin """)
			Else
				Call sendClient(theClient, "Error: invalid pin """)
			End If
			Call sendClient(theClient, value)
			Call sendClient(theClient, """<br>" & Chr(&H0d) & Chr(&H0a))

		Case "getpin"	' getpin=<pin number>
			pin = getPinNumber(value)
			If (pin <> 0) Then
				Dim b as Byte
				b = GetPin(pin)
				Call sendClient(theClient, "Pin """)
				Call sendClient(theClient, value)
				Call sendClient(theClient, """ is ")
				Call SendClient(theClient, IIf(CBool(b), "high", "low"))
				Call SendClient(theClient, "<br>" & Chr(&H0d) & Chr(&H0a))
			Else
				Call sendClient(theClient, "Error: invalid pin """)
				Call sendClient(theClient, value)
				Call sendClient(theClient, """<br>" & Chr(&H0d) & Chr(&H0a))
			End If

		Case Else		' unrecognized action name
				Call sendClient(theClient, "Error: unrecognized action """)
				Call sendClient(theClient, name)
				Call sendClient(theClient, """<br>" & Chr(&H0d) & Chr(&H0a))
		End Select
	End If
End Sub

'------------------------------------------------------------------------------
'
'' convertQueryEscapes
'
' Process a string potentially containing hexadecimal escapes of the form %XX
' where X represents a hexadecimal character.
'
Private Function convertQueryEscapes(ByVal str as String) as String
	convertQueryEscapes = ""
	Dim strIdx as Integer
	Dim sLen as Integer
	sLen = Len(str)
	strIdx = 1
	Do While (strIdx <= sLen)
		Dim c as Byte
		c = Asc(str, strIdx)
		If (c = Asc(" ")) Then
			' reached the end of the query string value
			Exit Do
		ElseIf (c = Asc("%")) Then
			If ((sLen - strIdx) >= 2) Then
				' found hexadecimal escape, append the equivalent value
				c = Shl(getHexVal(Asc(str, strIdx + 1)), 4) + getHexVal(Asc(str, strIdx + 2))
				convertQueryEscapes = convertQueryEscapes & Chr(c)
			strIdx = strIdx + 3
			Else
				' part of the hex escape is missing, truncate the query string value
				Exit Do
			End If
		Else
			' "normal" character, append to the string
			convertQueryEscapes = convertQueryEscapes & Chr(c)
			strIdx = strIdx + 1
		End If
	Loop
End Function

'------------------------------------------------------------------------------
'
'' getPinNumber
'
' Extract a pin number, specified in port.bit form or as a decimal number, from
' a string and return the representation of the pin number.  If the string does
' not contain a recognized pin number, contains a value that is too large, or
' contains unexpected characters, return zero.
'
Private Function getPinNumber(ByVal str as String) as Byte
	getPinNumber = 0
	Dim c as Byte
	Dim sLen as Integer
	sLen = Len(str)
	
	' check for port.pin form, e.g. A.3
	If ((sLen = 3) And (Asc(str, 2) = Asc("."))) Then
		c = Asc(str)
		If ((c >= Asc("a")) And (c <= Asc("z"))) Then
			' convert to upper case
			c = c - &H20
		End If
		If ((c >= Asc("A")) And (c <= Asc("Q"))) Then
			Dim port as Byte
			port = c - Asc("A")
			c = Asc(str, 3)
			If ((c >= Asc("0")) And (c <= Asc("7"))) Then
				getPinNumber = PortBit(port, c - Asc("0"))
			End If
		End If
	Else
		' assume the pin is given in decimal numeric form
		Dim pin as UnsignedInteger
		pin = 0
		Dim i as Integer

		For i = 1 to sLen
			c = Asc(str, i)
			If ((c >= Asc("0")) And (c <= Asc("9"))) Then
				pin = (pin * 10) + CUInt(c - Asc("0"))
				If (pin > 127) Then
					' invalid pin number value
					pin = 0
					Exit For
				End If
			Else
				' invalid character, abort
				pin = 0
				Exit For
			End If
		Next i
		getPinNumber = CByte(pin)
	End If
End Function

'------------------------------------------------------------------------------
'
'' sendClient
'
' Send the contents of a ZBasic string to the Ethernet client.  Because the
' string's characters could be in RAM, Program Memory or Persistent Memory, the
' characters are extracted one-by-one and sent to the client individually.
'
Private Sub sendClient(ByRef theClient as EthernetBasicClient, ByVal s as String)
	Dim i as Integer
	For i = 1 to Len(s)
		Call theClient.write(Asc(s, i))
	Next i
End Sub

'------------------------------------------------------------------------------
'
'' bufMatch
'
' This helper function determines if a string matches the characters at
' the beginning of a buffer.
'
Private Function bufMatch(ByRef buf() as Byte, ByVal s as String) as Boolean
	Dim i as Integer
	bufMatch = True
	For i = 1 to Len(s)
		If (buf(i) <> Asc(s, i)) Then
			bufMatch = False
			Exit For
		End If
	Next i
End Function

'------------------------------------------------------------------------------
'
'' FlashString
'
' This helper function converts a Program Memory address to a reference to
' a __FlashStringHelper used in the Arduino code.
'
Private Function FlashString(ByVal progMemAddr as Long) as __FlashStringHelper ByRef
	FlashString.DataAddress = progMemAddr
End Function

'------------------------------------------------------------------------------
'
'' dspIpAddr
'
' Output an IP address in dotted quad notation (e.g. 192.168.0.23) optionally
' followed by a newline.
'
Private Sub dspIpAddr(ByRef buf() as Byte, ByVal newline as Boolean = False)
	Dim i as Byte
	For i = 1 to 4
		If (i > 1) Then
			Debug.Print ".";
		End If
		Debug.Print buf(i);
	Next i
	If (newline) Then
		Debug.Print
	End If
End Sub

'------------------------------------------------------------------------------
'
'' getHexVal
'
' Return the value of a hexadecimal character .  If the byte is not a hex
' character, return 0.
'
Private Function getHexVal(ByVal c as Byte) as Byte
	If ((c >= Asc("0")) And (c <= Asc("9"))) Then
		getHexVal = c - Asc("0")
	ElseIf ((c >= Asc("a")) And (c <= Asc("f"))) Then
		getHexVal = c - (Asc("a") + 10)
	ElseIf ((c >= Asc("A")) And (c <= Asc("F"))) Then
		getHexVal = c - (Asc("A") + 10)
	Else
		getHexVal = 0
	End If
End Function

